--
-- Bitwise operations using algebra
--

-- references to common functions
local floor = math.floor

local bit = {}

-- max bit field width
-- 128 for float or 256 for double
--local maxb = math.log(math.huge)/math.log(2)
local maxb = 128
bit.maxb = maxb

-- max numeric value
local maxn = 2^(maxb - 1) - 1
bit.maxn = maxn

-- powers table cache
-- pows[n] = 2^n
local pows = {}
-- pows1[n] = 2(n - 1)
local pows1 = {}
-- pows2[n] = 2^n*2
local pows2 = {}
for i = 0, maxb - 1 do
  pows[i] = 2^i
  pows1[i] = 2^(i - 1)
  pows2[i] = pows[i]*2
end

-- Returns a bit field with only the 'b' flag set
function bit.new(b)
  return pows1[b]
end

-- Returns the b bit in the integer n
function bit.get(n, b)
  -- n%(2^b*2) >= 2^b
  if n%(pows2[b]) >= pows1[b] then
    return 1
  end
  return 0
end

-- Returns true if the b bit is 1 in the integer n
function bit.isset(n, b)
  -- n%(2^b*2) >= 2^b
  return n%(pows2[b]) >= pows1[b]
end

-- Sets the b bit in the integer n to 1
function bit.set(n, b)
  -- n%(2^b*2) < 2^b
  if n%(pows2[b]) < pows1[b] then
    n = n + pows[b]
  end
  return n
end

-- Sets the b bit in the integer n to 0
function bit.unset(n, b)
  -- n%(2^b*2) >= 2^b
  if n%(pows2[b]) >= pows1[b] then
    n = n - pows1[b]
  end
  return n
end

-- Left shifts the integer n by b bits
function bit.lshift(n, b)
  -- n*b^2
  return (n*pows[b])
end

-- Right shifts the integer n by b bits
function bit.rshift(n, b)
  -- floor(n/b^2)
  return floor(n/pows[b])
end

-- Tests two integers
function bit.test(n, n2)
  return n%(n2 + n2) >= n2
end

--[[
-- tests
assert(bit.test(0xff, 0xff))
assert(bit.test(0xff + 1, 0xff))
assert(bit.test(0xff, 0xff - 1))
assert(not bit.test(0xff - 1, 0xff))
assert(not bit.test(0xff, 0xff + 1))

assert(pows[0] == 2^0, pows[0])
assert(pows[1] == 2^1, pows[1])
assert(pows[2] == 2^2, pows[2])

assert(bit.lshift(1, 0) == 1)
assert(bit.lshift(1, 1) == 2)
assert(bit.lshift(1, 2) == 4)
assert(bit.lshift(1, 3) == 8)
assert(bit.lshift(1, 8) == 256)
assert(bit.lshift(0, 1) == 0)

assert(bit.rshift(0, 1) == 0)
assert(bit.rshift(256, 8) == 1)
assert(bit.rshift(8, 3) == 1)
assert(bit.rshift(4, 2) == 1)
assert(bit.rshift(2, 1) == 1)
assert(bit.rshift(1, 0) == 1)
assert(bit.rshift(1, 1) == 0)
    
local s = 0xff0000
s = bit.rshift(s, 0)
assert(s == 0xff0000, string.format("0x%x ~= 0xff0000", s))
s = bit.rshift(s, 8)
assert(s == 0x00ff00, string.format("0x%x ~= 0x00ff00", s))
s = bit.rshift(s, 8)
assert(s == 0x0000ff, string.format("0x%x ~= 0x0000ff", s))
s = bit.lshift(s, 8)
assert(s == 0x00ff00, string.format("0x%x ~= 0x00ff00", s))
s = bit.lshift(s, 8)
assert(s == 0xff0000, string.format("0x%x ~= 0xff0000", s))
s = bit.lshift(s, 0)
assert(s == 0xff0000, string.format("0x%x ~= 0xff0000", s))
local n = bit.new(0)
assert(bit.isset(n, 0), n)
n = bit.unset(n, 0)
assert(not bit.isset(n, 0), n)
local d = 0xff00ff

assert(bit.get(d, 0) == 1, string.format("0x%x bit %d is %d", d, 0, 1))
for i = 1, 8 do
  local b = bit.get(d, i)
  assert(b == 1, string.format("0x%x bit %d is %d", d, i, b))
end
for i = 9, 15 do
  local b = bit.get(d, i)
  assert(b == 0, string.format("0x%x bit %d is %d", d, i, b))
end
for i = 16, 24 do
  local b = bit.get(d, i)
  assert(b == 1, string.format("0x%x bit %d is %d", d, i, b))
end
assert(bit.get(d, 25) == 0, string.format("0x%x bit %d is %d", d, 25, 1))
]]

return bit