local insert = table.insert
local remove = table.remove
local concat = table.concat
local sort = table.sort

local format = string.format
local sub = string.sub
local gsub = string.gsub
local byte = string.byte
local match = string.match
local gmatch = string.gmatch
local len = string.len
local find = string.find

local floor = math.floor
local ceil = math.ceil
local max = math.max

--- Removes all values in a table
-- @param t Table
function table.clear(t)
  assert(t, "table is nil")
  for i in pairs(t) do
    t[i] = nil
  end
end

--- Counts the total number of elements in a table
-- @param t Table
-- @return Number of elements
function table.count(t)
  assert(t, "table is nil")
  local n = 0
  for _ in pairs(t) do
    n = n + 1
  end
  return n
end

--- Counts the number of duplicate values in a table
-- @param t Table
-- @return Number of duplicate values
local temp = {}
function table.dcount(t)
  assert(t, "table is nil")
  local d = 0
  for _, v in pairs(t) do
    if temp[v] then
      d = d + 1
    end
    temp[v] = true
  end
  table.clear(temp)
  return d
end

--- Reverses the elements in a list
-- @param t Table
-- @param r Resulting table (optional)
-- @return Reversed table
function table.reverse(t, r)
  assert(t, "table is nil")
  r = r or t
  local n = #t
  if t == r then
    for i = 1, floor(n/2) do
      local i2 = n - i + 1
      r[i], r[i2] = r[i2], r[i]
    end
  else
    for i, v in ipairs(t) do
      r[n - i + 1] = v
    end
  end
  return r
end

--- Finds the first occurrence of a value in a list
-- @param t Table
-- @param s Search value
-- @param o Starting index (optional)
-- @return Numeric index of the found value
function table.find(t, s, o)
  assert(t, "table is nil")
  o = o or 1
  for i = o, #t do
    if t[i] == s then
      return i
    end
  end
end

--- Finds the last occurrence of a value in a list
-- @param t Table
-- @param s Search value
-- @param o Starting index (optional)
-- @return Numeric index of the found value
function table.rfind(t, s, o)
  assert(t, "table is nil")
  o = o or #t
  -- iterate in reverse
  for i = o, 1, -1 do
    if t[i] == s then
      return i
    end
  end
end

--- Copies the contents from one table to another
--- Overwrites existing elements in the destination table
--- Preserves table cycles
-- @param s Source table
-- @param d Destination table (optional)
-- @param c Cache table (internal)
-- @return The resulting or destination table
local cache = {}
function table.copy(s, d, c)
  assert(s, "table is nil")
  assert(s ~= d, "source and destination tables must be different")
  d = d or {}
  -- clear the cache, initially
  if c == nil then
    c = cache
    table.clear(c)
  end
  -- copy elements from the source table
  for k, v in pairs(s) do
    if type(v) == "table" then
      if c[v] then
        -- reference cycle
        d[k] = c[v]
      else
        -- recursive copy
        if d[k] == nil then
          d[k] = {}
        end
        c[v] = d[k]
        table.copy(v, d[k], c)
      end
    else
      d[k] = v
    end
  end
  return d
end

--- Asserts that a given string is a valid concatenated key
-- @ck Concatenated key
function table.vckey(ck)
  -- must be a string
  local t = type(ck)
  if t ~= "string" then
    local e = format("key is not a string '%s'", tostring(ck))
    error(e)
  end
  -- must have length greater than 0
  local l = len(ck)
  if l == 0 then
    error("key is empty")
  end
  -- cannot end with the percent sign ('%')
  -- cannot end with the dot symbol ('.') except for ('%.')
  local lc = byte(ck, l)
  local nlc = byte(ck, l - 1)
  if (lc == 46 and nlc ~= 37) or (lc == 37) then
    local e = format("invalid termination character in key '%s'", ck)
    error(e)
  end
end

--- Accesses a table value using a concatenated key
-- @param t Table
-- @param ck Concatenated key
-- @return Value from table t
function table.get(t, ck)
  assert(t, "table is nil")
  table.vckey(ck)
  -- handle special sequence ('%.')
  ck = gsub(ck, "%%%.", "\001")
  for k, d in gmatch(ck, "([^%.]+)(%.?)") do
    k = gsub(k, "\001", ".")
    -- allow numeric keys access
    if t[k] == nil then
      local nk = tonumber(k)
      if nk and t[nk] then
        k = nk
      end
    end
    -- check if key exists
    if d == "." then
      local tk = type(t[k])
      if tk ~= "table" and tk ~= "userdata" then
        return
      end
    end
    -- descend to nested table or userdata
    t = t[k]
  end
  return t
end

--- Assigns the given value in a table using a concatenated key
--- Overwrites values if necessary
--- Creates sub-tables if necessary
-- @param t Table
-- @param ck Concatenated key
-- @param v Value
function table.set(t, ck, v)
  assert(t, "table is nil")
  table.vckey(ck)
  -- handle special sequence ('%.')
  ck = gsub(ck, "%%%.", "\001")
  for k, d in gmatch(ck, "([^%.]+)(%.?)") do
    k = gsub(k, "\001", ".")
    -- allow numeric keys access
    if t[k] == nil then
      local nk = tonumber(k)
      if t[nk] then
        k = nk
      end
    end
    if d == "." then
      -- create nested table if non-existent
      local tk = type(t[k])
      if tk ~= "table" and tk ~= "userdata" then
        t[k] = {}
      end
      -- descend to nested table or userdata
      t = t[k]
    else
      -- assign value
      t[k] = v
    end
  end
end

--- Returns a list of all concatenated keys in a table
--- Breadth first traversal provides the shortest concatenated keys
--- Ignores cycles
-- @param t Table
-- @param keys Output table (optional)
-- @return List of concatenated keys
local queue = {}
-- temporary lookup for visited tables
local visited = {}
local ckeys = {}
function table.gckeys(t, keys)
  assert(t, "table is nil")
  keys = keys or {}
  insert(queue, t)
  while #queue > 0 do
    -- dequeue
    local c = remove(queue, 1)
    -- queue child nodes
    visited[c] = true
    for k, v in pairs(c) do
      -- supported keys types: string or number
      local kt = type(k)
      if kt == 'string' or kt == 'number' then
        local ck = tostring(k)
        -- handle special sequence ('%.')
        ck = gsub(ck, "%.", "%%%.")
        local p = ckeys[c]
        if p then
          ck = format("%s.%s", p, ck)
        end
        insert(keys, ck)
        if type(v) == "table" then
          if not visited[v] then
            ckeys[v] = ck
            insert(queue, v)
          end
        end
      end
    end
  end
  -- optional, alphabetic sort
  -- todo: "1.10" is sorted above "1.2"
  sort(keys)
  -- clear the buffer tables
  table.clear(visited)
  table.clear(ckeys)
  return keys
end

--- Returns an iterator providing pairs of concatenated key and value
-- @param t Table
-- @return Iterator
function table.ckeys(t)
  assert(t, "table is nil")
  local keys = table.gckeys(t)
  local i = 0
  return function()
    i = i + 1
    if i > #keys then
      return
    end
    return keys[i], table.get(t, keys[i])
  end
end

local function toQutedString(sz)
  -- handle double quotes and new line
  -- backslash (hex 5C)
  sz = gsub(sz, '\\', '\\\\')
  -- double quote character (hex 22)
  sz = gsub(sz, '"', '\\\"')
  -- newline (hex 0A)
  sz = gsub(sz, '\n', '\\n')
  -- carriage return (hex 0D) becomes \r
  sz = gsub(sz, '\r', '\\r')
  -- todo: nul character (hex 00) becomes \000
  sz = format('"%s"', sz)
  -- todo: format "%q" is a little funny about new lines
  --sz = format('%q', sz)
  return sz
end

--- Converts table to a string representation
--- Makes the string as compact as possible
--- Handles boolean and special keywords as keys
-- @param t Table
-- @return String
local keywords =
{
  ["and"] = true, ["break"] = true, ["do"] = true,
  ["else"] = true, ["elseif"] = true, ["end"] = true,
  ["false"] = true, ["for"] = true, ["function"] = true,
  ["if"] = true, ["in"] = true, ["local"] = true,
  ["nil"] = true, ["not"] = true, ["or"] = true,
  ["repeat"] = true, ["return"] = true, ["then"] = true,
  ["true"] = true, ["until"] = true, ["while"] = true
}
local infinities =
{
  [tostring(1/0)] = '1/0', -- 1.#INF or math.huge
  [tostring(-1/0)] = '-1/0', -- -1.#INF or -math.huge
  [tostring(0/0)] = '0/0' -- 1.#IND or NaN
}
function table.tostring(t, c, i)
  -- todo: slow and creates a lot of intermediate strings
  -- todo: non-sorted for non-numeric keys
  -- todo: does not handle cycles
  -- todo: tables, functions or userdata keys are converted using "tostring"
  i = i or 0
  assert(t, "table is nil")
  local t2 = {}
  -- is it a list?
  local n = table.count(t) == #t
  for k, v in pairs(t) do
    -- value
    local vt = type(v)
    if vt == "table" then
      v = table.tostring(v, c, i + 1)
    elseif vt == "string" then
      v = toQutedString(v)
    elseif vt == "number" then
      v = tostring(v)
      v = infinities[v] or v
    else
      v = tostring(v)
    end
    local pair = v
    if n == false then
      -- key
      local kt = type(k)
      if kt == "string" then
        if keywords[k] or not match(k, '^[%a_][%w_]*$') then
          k = toQutedString(k)
          k = format('[%s]', k)
        end
      elseif kt == "number" then
        k = tostring(k)
        k = infinities[k] or k
        k = format('[%s]', k)
      else
        k = tostring(k)
        k = format('[%s]', k)
      end
      -- key and value pair
      pair = format("%s=%s", k, v)
    end
    -- output
    insert(t2, pair)
  end
  if not c and #t2 > 0 then
    local f1 = string.rep("  ", i)
    local f2 = string.rep("  ", i + 1)
    local tsz = concat(t2, ",\n" .. f2)
    return format("\n%s{\n%s%s\n%s}", f1, f2, tsz, f1)
  else
    local tsz = concat(t2, ",")
    return format("{%s}", tsz)
  end
end

--- Converts string to a table
-- @param sz String
-- @return Table
function table.fromstring(sz)
  assert(type(sz) == "string", "input is not a string")
  -- todo: highly unsafe
  local f, e = loadstring("return " .. sz)
  if e then
    error(e)
  end
  return f()
end

--- Converts list to a CSV representation
--- Works only with numerically indexed tables
--- "nil" values are included too, therefore:
--- table.tocsv({[5] = true}) ==> ",,,,true"
-- @param t Table
-- @param c Number of columns (optional)
-- @return String
function table.tocsv(t, c)
  assert(t, "table is nil")
  local t2 = {}
  local i2 = 0
  for i, v in pairs(t) do
    if type(i) == "number" then
      if floor(i) == i and i > 0 then
        i2 = max(i2, i)
      end
    end
  end
  c = c or i2
  assert(c > 0, "number of columns is zero")
  -- number of rows
  local r = ceil(i2/c)
  for i = 1, i2 do
    local v = t[i]
    local vt = type(v)
    if vt == "string" then
      -- handle double quotes and new line
      v = toQutedString(v)
    elseif vt == "nil" then
      v = ""
    else
      v = tostring(v)
    end
    insert(t2, v)
    if i < r*c then
      if c > 0 and i%c == 0 then
        insert(t2, "\n")
      else
        insert(t2, ",")
      end
    end
  end
  for i = 1, r*c - i2 - 1 do
    insert(t2, ",")
  end
  insert(t2, "\n")
  return concat(t2)
end

--- Converts a CSV string to a table
--- Adapted from: http://lua-users.org/wiki/CsvUtils
-- @param sz String
-- @return Table
function table.fromcsv(sz)
  assert(type(sz) == "string", "input is not a string")
  local t = {}
  local tn = 0
  local si, ei = 1, len(sz)
  while si < ei do
    -- single line
    local nl = find(sz, "\n", si, true) or ei
    local line = sub(sz, si, nl)
    si = nl + 1

    -- ending comma
    line = line .. ','
    local fs = 1
    repeat
      local f = nil
      if find(line, '^"', fs) then
        -- quoted
        local a, c
        local i = fs
        repeat
          -- find closing quote
          a, i, c = find(line, '"("?)', i + 1)
          -- quote not followed by quote?
        until c ~= '"'
        assert(i, 'unmatched double quote')
        f = sub(line, fs + 1, i - 1)
        f = gsub(f, '""', '"')
        fs = find(line, ',', i, true) + 1
      else
        -- unquoted
        -- find next comma
        local nexti = find(line, ',', fs, true)
        f = sub(line, fs, nexti - 1)
        if f == "" then
          f = nil
        elseif f == "true" then
          f = true
        elseif f == "false" then
          f = false
        else
          f = tonumber(f) or f
        end
        fs = nexti + 1
      end
      tn = tn + 1
      t[tn] = f
    until fs > len(line)
  end
  return t
end

local function toXMLString(sz)
  sz = gsub(sz, "&", "&amp;")
  sz = gsub(sz, "<", "&lt;")
  sz = gsub(sz, ">", "&gt;")
  --sz = gsub(sz, "'", "&apos;")
  sz = gsub(sz, "\"", "&quot;")
  -- replace non printable char -> "&#xD;"
  sz = gsub(sz, "([^%w%&%;%p%\t% ])",
  function (c) 
    return format("&#x%X;", byte(c)) 
    --return format("&#x%02X;", byte(c)) 
    --return format("&#%02d;", byte(c)) 
  end)
  return sz
end

local function fromXMLString(sz)
  sz = gsub(sz, "&#x([%x]+)%;",
  function(h) 
    return char(tonumber(h,16)) 
  end)
  sz = gsub(sz, "&#([0-9]+)%;",
  function(h) 
    return char(tonumber(h,10)) 
  end)
	sz = gsub(sz, "&quot;", "\"")
	sz = gsub(sz, "&apos;", "'")
	sz = gsub(sz, "&gt;", ">")
	sz = gsub(sz, "&lt;", "<")
	sz = gsub(sz, "&amp;", "&")
	return sz
end

--- Adapted from Roberto Ierusalimschy: http://lua-users.org/wiki/LuaXml
function table.fromxml(s)
  assert(type(s) == "string", "input is not a string")
  -- replace prologue
  s = gsub(s, "<%?(.*)%?>", "")
  -- replace doctype
  --s = gsub(s, "<!(.*)%>", "")
  -- replace comments
  s = gsub(s, "<!%-%-(.-)%-%->", "")
  
  local stack = {}
  local top = {}
  insert(stack, top)
  local i = 1
  while true do
    local ni, j, c, tag, xarg, empty = find(s, "<(%/?)([%w:]+)(.-)(%/?)>", i)
    if not ni then
      break
    end
    local text = sub(s, i, ni - 1)
    if not find(text, "^%s*$") then
      -- replace comments
      -- todo: replace spaces around comments
      --text = gsub(text, "<!--(.*)-->", "")
      insert(top, text)
    end
    if empty == "/" or c == "" then
      -- start tag
      -- store the tag name in the zero index [0]
      local t = { [0] = tag }
      -- load attributes
      gsub(xarg, "([%-%w]+)=([\"'])(.-)%2",
      function (w, _, a)
        t[w] = fromXMLString(a)
      end)
      if empty == "/" then
        -- empty
        insert(top, t)
      else
        -- non-empty
        top = t
        -- new level
        insert(stack, top)
      end
    else
      -- end tag
      local toclose = remove(stack)
      -- remove top
      top = stack[#stack]
      if #stack < 1 then
        error("nothing to close with " .. tag)
      end
      if toclose[0] ~= tag then
        error("trying to close " .. toclose[0] .. " with " .. tag)
      end
      insert(top, toclose)
    end
    i = j + 1
  end
  local text = sub(s, i)
  if not find(text, "^%s*$") then
    insert(stack[#stack], text)
  end
  if #stack > 1 then
    error("unclosed " .. stack[#stack][0])
  end
  return stack[1]
end

function table.toxml(t)
  local sz = ""
  
  return sz
end

--- Adapted from Bart van Strien: inifile
function table.toini(t)
	local sz = ""
	for section, s in pairs(t) do
		sz = sz .. format("[%s]\n", section)
    if type(s) == "table" then
      for key, value in pairs(s) do
        value = tostring(value)
        sz = sz .. format("%s=%s\n", key, value)
      end
      sz = sz .. "\n"
    end
	end
	return sz
end

function table.fromini(s)
  assert(type(s) == "string", "input is not a string")
	local t = {}
	local section
	for line in lines(name) do
    local s2 = match(line, "^%[([^%]]+)%]$")
		if s2 then
			section = s2
			t[section] = t[section] or {}
		end
		local key, value = match(line, "^([%w_]+)%s-=%s-(.+)$")
		if section and key and value then
			if value == "true" then
        value = true
      elseif value == "false" then
        value = false
      else
        value = tonumber(value) or value
      end
			t[section][key] = value
		end
  end
  return t
end